##TODO: Label the columns
##TODO: Collapse same text
##TODO: Interline differences should be marked with a new color
##TODO: Pod documentation

# The lines from here to the =cut are part of Perl's
# internal documentation system.  If you want 
# view the documentation use the command:
#
#	pod2text <script>
#
=pod

=head1 NAME

pdiff - A diff/merge program

=head1 SYNOPSIS

    perl pdiff.pl -a <ancestor> [-o <out_file>] [-n] [-D] file1 file2

=head1 DESCRIPTION

The I<pdiff.pl> command perform a graphical three way diff between two files and
a common ancestor.   The user can then (optionally) merge the text into one
resulting file.

=head1 OPTIONS

The I<pdiff.pl> takes the following arguments:

=over 4

=item B<-a> I<ancestor>

The file that was modified to produce the two versions: I<file1> 
and I<file2>. (Required.)

=item B<-o> I<out_file>

File that is to get the merged output.

=item B<-n>

Don't include an ancestor column in the display.

=item B<-D>

Turn on debug.  (Currently does nothing.)

=back

=head1 BUGS/TODO

=head2 BUGS

There should be a way of diffing just two files.

=head2 TODO

Label the columns

Collapse same text

Interline differences should be marked with a new color

=head1 REQUIREMENTS

Requires a I<diff3> program like the one that comes with the GNU
diff utilities.

=head1 AUTHOR

Steve Oualline, E<lt>oualline@www.oualline.comE<gt>.

=head1 COPYRIGHT

Copyright 2002 Steve Oualline.
This program is distributed under the GPL.  

=cut
use strict;
use warnings;
use Getopt::Std;
use Text::Tabs;
use Tk;
use Tk::ROText;
use Tk::Adjuster;
use Tk::Balloon;
use vars qw($opt_o $opt_a $opt_D $opt_n);
#
# diff_info -- Information about the differences for each of the 
#	four windows (original, file1, file2, merge)
#
# Information:
#
#	mark name => diff.start.<index>	First place in text where the diff is
#					(Tk text mark, not an item in diff_info)
#
#	show{original} =>		What we see
#	    line => <line_number>	Line number where the change starts
#				        (File line number, not display line number)
#	    diff => <lines>		Number of lines different
#	    pad => <lines>		Number of padding lines
#
#	show{file1} =>			Same for file 1 window
#	show{file2} =>			Same for file 2 window
#	show{merge} =>			Same for merge result
#
#	result => <result_type>		Which result we choose for
#					the merge
#
# <result_type> = {none, left, right, left_right, right_left}
#
my @diff_info = ();	# Information on the differences
my $cur_diff = 0;	# Index of the current difference

my %lines = ();		# Lines indexed by file type
my %tk_text = (); 	# Text boxes by file type
my %colors = (
	original => 'green',
	file1    => 'lightblue',
	file2    => 'yellow',
	merge    => 'red'
);

#
# File types:
#	original, file1, file2, merge

$tabstop = 8;		# Number of spaces in a tab

my $diff3_cmd = "diff3";

my $original;	# Original file
my $file1;	# First edit
my $file2;	# Second edit
my $merge;	# Merge file

my $tk_scroll;	# The scrollbar
my $tk_mw;	# The main window

my $pad = "~\n";	# Line to use for padding


################################################################
# push_change(\@change) -- Add a change to the diff_list
################################################################
sub push_change($)
{
    my $change = shift;	# The change information
    if (not defined($$change[1]{line}))
    {
	return;
    }
    my $max_size = $$change[1]{size};	# largest change
    if ($$change[2]{size} > $max_size)
    {
	$max_size = $$change[2]{size};
    }
    if ($$change[3]{size} > $max_size)
    {
	$max_size = $$change[3]{size};
    }
    # Information about the diff
    my %diff = (
	show => {
	    original => {
		line => $$change[1]{line},
		diff => $$change[1]{size},
		pad => $max_size - $$change[1]{size}
	    },
	    file1 => {
		line => $$change[2]{line},
		diff => $$change[2]{size},
		pad => $max_size - $$change[2]{size}
	    },
	    file2 => {
		line => $$change[3]{line},
		diff => $$change[3]{size},
		pad => $max_size - $$change[3]{size}
	    },
	    merge => {
		line => $$change[1]{line},
		diff => 0,
		pad => $max_size
	    },
	},
	result => "none"
    );
    # Save the result
    push(@diff_info, {%diff});
}
################################################################
# compute_diff_info -- Compute the difference information
################################################################
sub compute_diff_info()
{
    open DIFF_FILE, "$diff3_cmd $original $file1 $file2|" or
       die ("Could not execute the difference command");

    my @change;	# Change information
    # $change[1] -- Change for file 1
    # $change[2] -- Change for file 3
    # $change[3] -- Change for file 3
    # $change[x]{line} -- Line number where the change occurs
    # $change[x]{size} -- The number of lines affected

    while (<DIFF_FILE>)
    {
	if (/^===/)
	{
	    push_change(\@change);
	}

	# We care only for the numbered lines
	if (! /^\d:/)
	{
	    next;
	}
	if (/^(\d):(\d*),(\d*)(.)/)
	{
	    $change[$1]{line} = $2-1;
	    $change[$1]{size} = $3 - $2 + 1;
	    if ($4 ne 'c')
	    {
		die ("Unknown code: $4");
	    }
	    next;
	}
	if (/^(\d):(\d*)(.)\n/)
	{
	    if ($3 eq 'c')
	    {
		$change[$1]{line} = $2-1;
		$change[$1]{size} = 1;
	    }
	    elsif ($3 eq 'a')
	    {
		$change[$1]{line} = $2;
		$change[$1]{size} = 0;
	    }
	    next;
	}
	else
	{
	    die("Unknown code: $3");
	}
    }
    push_change(\@change);
}
################################################################
# read_file($file_name) -- Read a file and return the lines in it
#
# Returns a list of the lines read
################################################################
sub read_file($)
{
    my $file_name = shift;	# The file name

    open IN_FILE, "<$file_name" or
        die("Unable to open $file_name for reading");

    # The lines from the file
    my @lines = <IN_FILE>;

    close(IN_FILE);
    return (@lines);
}
################################################################
# display_text_diff -- Display the diffs as text
# 	(DEBUG routine)
################################################################
sub display_text_diff()
{

    my $cur_diff = 0;	# The current diff
    my $original_lc = 0;# Original file line counter
    my $file1_lc = 0;	# File1 file line counter
    my $file2_lc = 0;	# File2 file line counter

    my @original_lines = @{$lines{original}};
    my @file1_lines = @{$lines{file1}};
    my @file2_lines = @{$lines{file2}};
    chomp(@original_lines);
    chomp(@file1_lines);
    chomp(@file2_lines);

BIG_LOOP:
    while ($original_lc <= $#original_lines)
    {
	my $next_change;	# Location of the next change
	if (not defined($diff_info[$cur_diff]))
	{
	    $next_change = 999999;	# No next change
	}
	else
	{
	    $next_change = $diff_info[$cur_diff]->{show}{original}{line};
	}

	while ($original_lc < $next_change)
	{
	    if ($original_lc > $#original_lines)
	    {
		last BIG_LOOP;
	    }
	    printf "S: %-25s | %-25s | %-25s\n",
	    	$original_lines[$original_lc],
	    	$file1_lines[$file1_lc],
	    	$file2_lines[$file2_lc];
	    $original_lc++;
	    $file1_lc++;
	    $file2_lc++;
	}
	if (defined($diff_info[$cur_diff]))
	{
	    my $diff_size = 
		$diff_info[$cur_diff]->{show}{original}{diff} +
		$diff_info[$cur_diff]->{show}{original}{pad};

	    for (my $cur_line = 0; $cur_line < $diff_size; $cur_line++)
	    {
		my $original_str = "";
		my $file1_str = "";
		my $file2_str = "";

		if ($cur_line < $diff_info[$cur_diff]->{show}{original}{diff})
		{
		    $original_str = $original_lines[$original_lc];
		    $original_lc++;
		}
		if ($cur_line < $diff_info[$cur_diff]->{show}{file1}{diff})
		{
		    $file1_str = $file1_lines[$file1_lc];
		    $file1_lc++;
		}
		if ($cur_line < $diff_info[$cur_diff]->{show}{file2}{diff})
		{
		    $file2_str = $file2_lines[$file2_lc];
		    $file2_lc++;
		}
		printf "D: %-25s | %-25s | %-25s\n",
		    $original_str, $file1_str, $file2_str;
	    }
	    $cur_diff++;
	}
    }
}
############################################################
# error_msg($message) -- Display an error message
############################################################
sub error_msg($)
{
    my $message = shift; # The message to display 

    $tk_mw->messageBox(
	    -icon => "error",
	    -type => "OK",
	    -title => "Error",
	    -message => $message,
	    -background => "pink"
	);
    return;
}
################################################################
# write_out_file -- Write the output file
################################################################
sub write_out_file()
{

    if (not defined($merge)) {
        return;
    }
    my $cur_diff = 0;	# The current diff
    my $original_lc = 0;# Original file line counter
    my $file1_lc = 0;	# File1 file line counter
    my $file2_lc = 0;	# File2 file line counter

    # A few variables to make our life easier
    my $original_lines = \@{$lines{original}};
    my $file1_lines = \@{$lines{file1}};
    my $file2_lines = \@{$lines{file2}};

    open OUT_FILE, ">$merge" or
	return (error_msg("Unable to open the output file: $merge"));
BIG_LOOP:
    while ($original_lc <= $#$original_lines)
    {
	my $next_change;	# Location of the next change
	if (not defined($diff_info[$cur_diff]))
	{
	    $next_change = 999999;	# No next change
	}
	else
	{
	    $next_change = $diff_info[$cur_diff]->{show}{original}{line};
	}

	while ($original_lc < $next_change)
	{
	    if ($original_lc > $#$original_lines)
	    {
		last BIG_LOOP;
	    }
	    print OUT_FILE $$original_lines[$original_lc];
	    $original_lc++;
	    $file1_lc++;
	    $file2_lc++;
	}
	# We just found a difference.   Decide what to do with it
	if (defined($diff_info[$cur_diff]))
	{
	    # The difference type
	    my $diff_type = $diff_info[$cur_diff]{result};

	    if ($diff_type eq 'left')
	    {
		# The number of lines in this section
		my $limit = $diff_info[$cur_diff]{show}{file1}{diff};
		for (my $index = 0; $index < $limit; $index++)
		{
		    print OUT_FILE $$file1_lines[$index + $file1_lc];
		}
	    }
	    elsif ($diff_type eq 'right')
	    {
		# The number of lines in this section
		my $limit = $diff_info[$cur_diff]{show}{file2}{diff};
		for (my $index = 0; $index < $limit; $index++)
		{
		    print OUT_FILE $$file2_lines[$index + $file2_lc];
		}
	    }
	    elsif ($diff_type eq 'left_right')
	    {
		# The number of lines in this section
		my $limit = $diff_info[$cur_diff]{show}{file1}{diff};
		for (my $index = 0; $index < $limit; $index++)
		{
		    print OUT_FILE $$file1_lines[$index + $file1_lc];
		}

		# The number of lines in this section
		$limit = $diff_info[$cur_diff]{show}{file2}{diff};
		for (my $index = 0; $index < $limit; $index++)
		{
		    print OUT_FILE $$file2_lines[$index + $file2_lc];
		}
	    }
	    elsif ($diff_type eq 'right_left')
	    {
		# The number of lines in this section
		my $limit = $diff_info[$cur_diff]{show}{file2}{diff};
		for (my $index = 0; $index < $limit; $index++)
		{
		    print OUT_FILE $$file1_lines[$index + $file2_lc];
		}

		# The number of lines in this section
		$limit = $diff_info[$cur_diff]{show}{file1}{diff};
		for (my $index = 0; $index < $limit; $index++)
		{
		    print OUT_FILE $$file1_lines[$index + $file1_lc];
		}
	    }
	    else
	    {
		die("Unknown diff type $diff_type");
	    }
	    $original_lc += $diff_info[$cur_diff]{show}{original}{diff};
	    $file1_lc += $diff_info[$cur_diff]{show}{file1}{diff};
	    $file2_lc += $diff_info[$cur_diff]{show}{file2}{diff};
	    $cur_diff++;
	}
    }
    close OUT_FILE;
}
################################################################
# total_diff($cur_file) -- Return the total number 
# of lines in a differece block
################################################################
sub total_diff($)
{
    my $cur_file = shift;	# the file to look at
    return (
	$diff_info[$cur_diff]{show}{$cur_file}{diff} +
	$diff_info[$cur_diff]{show}{$cur_file}{pad}
    );
}
################################################################
# compute_diff_loc($type, $cur_diff) -- Return the location of 
#	the next place to put a diff block
################################################################
sub compute_diff_loc($$)
{
    my $type = shift;
    my $cur_diff = shift;

    if (defined($diff_info[$cur_diff]))
    {
	return ($diff_info[$cur_diff]{show}{$type}{line});
    }
    else
    {
	return (9999999);
    }
}
################################################################
# fill_box($type) -- Fill in the text box with text
################################################################
sub fill_box($)
{
    my $type = shift;	# File type

    my $file_lines = $lines{$type};	# The lines in the file
    my $tk_text_box = $tk_text{$type};	# The text box
    my $cur_diff = 0;	# Current diff number

    # Next place to put padding
    my $next_pad = compute_diff_loc($type, $cur_diff);	

    #
    # diff_start[x]	-- Line diff starts on
    #
    my @diff_start = ();	# hash containing location information

    for (my $line_number = 0; $line_number <= $#$file_lines; $line_number++)
    {
	if ($line_number == $next_pad)
	{
	    my $diff_lines = $diff_info[$cur_diff]{show}{$type}{diff};
	    my $pad_lines = $diff_info[$cur_diff]{show}{$type}{pad};

	    #
	    # Create a mark for the start of this diff
	    #
	    $diff_start[$cur_diff] = $tk_text_box->index('end -1 lines');

	    for (my $cur_line = 0; $cur_line < $diff_lines; $cur_line++)
	    {
		$tk_text_box->insert('end', expand($$file_lines[$line_number]));
		$line_number++;
	    }
	    $line_number--;

	    for (my $cur_line = 0; $cur_line < $pad_lines; $cur_line++)
	    {
		$tk_text_box->insert('end', $pad);
	    }
	    $cur_diff++;
	    $next_pad = compute_diff_loc($type, $cur_diff);
	}
	else
	{
	    $tk_text_box->insert('end', expand($$file_lines[$line_number]));
	}
    }
    for ($cur_diff = 0; $cur_diff <= $#diff_start; $cur_diff++)
    {
	$tk_text_box->markSet("diff.start.$cur_diff", $diff_start[$cur_diff]);
	$tk_text_box->markGravity("diff.start.$cur_diff", 'left');

	if ($type eq "merge")
	{
	    # The the type of the difference
	    my $diff_type = $diff_info[$cur_diff]{result};

	    # The total number of lines in the diff
	    my $total = total_diff("merge");

	    if ($diff_type eq "left")
	    {
		$tk_text_box->tagAdd('diff_file1', $diff_start[$cur_diff],
			"$diff_start[$cur_diff] + $total lines");
	    } 
	    elsif ($diff_type eq "right")
	    {
		$tk_text_box->tagAdd('diff_file2', $diff_start[$cur_diff],
			"$diff_start[$cur_diff] + $total lines");
	    }
	    else
	    {
		die("Unknown diff type $diff_type");
	    }
	}
	else
	{
	    # The total number of lines in the diff
	    my $total = total_diff($type);

	    $tk_text_box->tagAdd('diff', $diff_start[$cur_diff],
		    "$diff_start[$cur_diff] + $total lines");
	}
    }
}
    

################################################################
# scroll_all -- Scroll all the text screens
################################################################
sub scroll_all($$@)
{
    my $tk_scroll = shift;	# The scrollbar
    my $cur_text = shift;	# Text widget that's scrolling
    my @args = @_;		# The scroll arguements

    $tk_scroll->set(@args);
    my ($top, $bottom) = $cur_text->yview;
    foreach my $cur_box (values %tk_text)
    {
	$cur_box->yviewMoveto($top);
    }
}
################################################################
# set_diff -- Set the difference number
################################################################
sub set_diff($)
{
    $cur_diff = shift;	# Get the new current differnce

    if ($cur_diff < 0)
    {
	$cur_diff = 0;
    }
    if ($cur_diff > $#diff_info)
    {
	$cur_diff = $#diff_info;
    }
    foreach my $cur_file (keys %tk_text)
    {
	my %mark_hash = map +($_, 1), $tk_text{$cur_file}->markNames();

	if (not defined($mark_hash{"diff.start.$cur_diff"}))
	{
	    print "### Missing mark diff.start.$cur_diff\n";
	    next;
	}
	while (1)
	{
            # The start / end of the current range for this tag
	    my ($start, $end) = 
		$tk_text{$cur_file}->tagNextrange('cur_diff', "0.0");

	    if (not defined ($start))
	    {
		last;
	    }
	    $tk_text{$cur_file}->tagRemove('cur_diff', $start, $end);
	}

	# The total number of lines in the diff block
	my $total = total_diff($cur_file);

	# The start of the different
	my $start = $tk_text{$cur_file}->index("diff.start.$cur_diff");

	# The end of the difference (including padding)
	my $end = $tk_text{$cur_file}->index("diff.start.$cur_diff + $total lines");

	$tk_text{$cur_file}->tagAdd('cur_diff', $start, $end);
	$tk_text{$cur_file}->see($start);
    }
}
################################################################
# cb_exit -- Called to write the file and exit
################################################################
sub cb_exit()
{
    write_out_file();
    exit (0);
}
################################################################
# cb_abort -- Exit immediately
################################################################
sub cb_abort() 
{
    exit (8);
}

################################################################
# adjust_padding -- Adjust the padding for the current diff
################################################################
sub adjust_padding()
{
    my $big_diff = 0;		# The number of lines we need to display
    my %file_total;		# Total lines per file

    foreach my $cur_file (qw(file1 file2 original merge))
    {
	if ($diff_info[$cur_diff]{show}{$cur_file}{diff} > $big_diff)
	{
	    $big_diff = $diff_info[$cur_diff]{show}{$cur_file}{diff};
	}
    }
    # Now adjust the padding
    foreach my $cur_file (keys %tk_text)
    {
	while (total_diff($cur_file) < $big_diff)
	{
	    # The total diff size
	    my $total = total_diff($cur_file);

	    $tk_text{$cur_file}->insert(
		    "diff.start.$cur_diff + $total lines",
		    $pad);
	    $diff_info[$cur_diff]{show}{$cur_file}{pad}++;
	}
	while (total_diff($cur_file) > $big_diff)
	{
	    # The total diff size
	    my $diff = $diff_info[$cur_diff]{show}{$cur_file}{diff};

	    $tk_text{$cur_file}->delete(
		    "diff.start.$cur_diff + $diff lines",
		    "diff.start.$cur_diff + $diff lines lineend + 1 chars");
	    $diff_info[$cur_diff]{show}{$cur_file}{pad}--;
	}
    }
}

################################################################
# clear_tags -- Clear the tags for the merge window's current
#	diff.
################################################################
sub clear_tags()
{
    foreach my $cur_file (keys %tk_text)
    {
	# The total number of lines in the diff block
	my $total = total_diff($cur_file);
	$tk_text{merge}->tagRemove("diff_$cur_file",
		"diff.start.$cur_diff",
		"diff.start.$cur_diff + $total lines");
    }
    foreach my $cur_file (keys %tk_text)
    {
	if ($cur_file eq "merge")
	{
	    next;	# Merge is taken care of already
	}
	# The total number of lines in the diff block
	my $total = total_diff($cur_file);

	$tk_text{$cur_file}->tagRemove("diff",
		"diff.start.$cur_diff",
		"diff.start.$cur_diff + $total lines");
    }
}
################################################################
# set_tags -- Set the tags for the file window's current
#	diff. (all except the merge window)
################################################################
sub set_tags()
{
    foreach my $cur_file (keys %tk_text)
    {
	if ($cur_file eq "merge")
	{
	    next;	# Merge is taken care of outside
	}
	# The total number of lines in the diff block
	my $total = total_diff($cur_file);
	$tk_text{$cur_file}->tagAdd("diff",
		"diff.start.$cur_diff",
		"diff.start.$cur_diff + $total lines");
    }
}
################################################################
# left_right($flag, $file) -- Select left or right
################################################################
sub left_right($$)
{
    my $flag = shift;	# The flag to put in the result field
    my $file = shift;	# The file to get the text from

    clear_tags();

    # Total number of lines in the merge block
    my $total = total_diff("merge");

    $tk_text{merge}->delete(
	    "diff.start.$cur_diff",
	    "diff.start.$cur_diff + $total lines");

    my $text = $tk_text{$file}->get(
	    "diff.start.$cur_diff",
	    "diff.start.$cur_diff + $diff_info[$cur_diff]{show}{$file}{diff} lines");

    $tk_text{merge}->insert(
	    "diff.start.$cur_diff",
	    $text);

    $diff_info[$cur_diff]{show}{merge}{diff} = 
	$diff_info[$cur_diff]{show}{$file}{diff};

    $diff_info[$cur_diff]{show}{merge}{pad} = 0;
    $diff_info[$cur_diff]{result} = $flag;

    adjust_padding();

    $total = total_diff("merge");
    $tk_text{merge}->tagAdd("diff_$file",
	    "diff.start.$cur_diff",
	    "diff.start.$cur_diff + $total lines");

    set_tags();
    set_diff($cur_diff);
}

################################################################
# cb_use_left -- Select the text on the left
################################################################
sub cb_use_left() 
{
    left_right('left', 'file1');
}
################################################################
# cb_use_right -- Select the text on the right
################################################################
sub cb_use_right() 
{
    left_right('right', 'file2');
}
################################################################
# then_insert($flag, $first, $send) 
#	Insert x then y
################################################################
sub then_insert($$$)
{
    my $flag = shift;		# The result flag
    my $first_file = shift;	# The first file to insert
    my $second_file = shift;	# Second file to insert

    clear_tags();
    # Total number of lines in the merge block
    my $total = total_diff("merge");

    $tk_text{merge}->delete(
	    "diff.start.$cur_diff",
	    "diff.start.$cur_diff + $total lines");

    # The # lines different in the first file
    my $first_diff = $diff_info[$cur_diff]{show}{$first_file}{diff};

    # The text from the first file
    my $first_text = $tk_text{$first_file}->get(
	    "diff.start.$cur_diff",
	    "diff.start.$cur_diff + $first_diff lines");

    # The # lines different in the second file
    my $second_diff = $diff_info[$cur_diff]{show}{$second_file}{diff};

    # The text from the second file
    my $second_text = $tk_text{$second_file}->get(
	    "diff.start.$cur_diff",
	    "diff.start.$cur_diff + $second_diff lines");

    $tk_text{merge}->insert(
	    "diff.start.$cur_diff",
	    $first_text.$second_text);

    $diff_info[$cur_diff]{show}{merge}{diff} = $first_diff + $second_diff;
    $diff_info[$cur_diff]{show}{merge}{pad} = 0;
    $diff_info[$cur_diff]{result} = $flag;

    adjust_padding();

    $tk_text{merge}->tagAdd("diff_$first_file",
	    "diff.start.$cur_diff",
	    "diff.start.$cur_diff + $first_diff lines");

    $total = total_diff("merge");

    $tk_text{merge}->tagAdd("diff_$second_file",
	    "diff.start.$cur_diff + $first_diff lines",
	    "diff.start.$cur_diff + $total lines");

    set_tags();
    set_diff($cur_diff);
}

################################################################
# cb_use_left_right -- Insert left part, then right
################################################################
sub cb_use_left_right() 
{
    then_insert("left_right", "file1", "file2");
}
################################################################
# cb_use_left_right -- Insert left part, then right
################################################################
sub cb_use_right_left() 
{
    then_insert("right_left", "file2", "file1");
}

################################################################
# cb_click -- called when something is clicked.
################################################################
sub cb_click
{
    # The location in screen space of the line
    my $loc = $_[0]->index('current');
    $loc = int($loc);	# Drop character index
    
    # Now transform the location into an index into the diff_info array
    my $cur_line = 0;	# Current line number
    my $pad_total = 0;	# padding so far
    my $index = 0;	# Index into the diff info array

    while (($index < $#diff_info) and ($cur_line < $loc))
    {
	# The information about this diff (which file used does not matter)
	my $show_info = $diff_info[$index]{show}{file1};
	$pad_total += $show_info->{pad};
	$cur_line = $show_info->{line} + $pad_total + $show_info->{diff};
	if ($cur_line >= $loc)
	{
	    last;
	}
	$index++;
    }
    set_diff($index); 
}
################################################################
# create_select_buttons($frame) -- Create the buttons to 
# select diffs and move things around
################################################################
sub create_select_buttons($)
{
    # The frame with the actions in it
    my $tk_action_frame = shift;	
    my $width = 14;
    my %help = (
	    "Next" =>	"Go to  the next difference",
	    "Previous" => "Go to  the previous difference",
	    "First" => "Go to  the first difference",
	    "Last" => "Go to  the last difference",
	    "Left" => "Put the changes from the left file in the output file",
	    "Right" => "Put the changes from the right file in the output file",
	    "Left then Right" => "Put the changes from both files in the output file\n".
	    			"Insert the Left file changes followed by the Right ones",
	    "Right then Left" => "Put the changes from both files in the output file\n".
	    			"Insert the Right file changes followed by the Left ones");

    # The button to select the next diff
    $tk_action_frame->Button(
	    -text => "Next",
	    -command => sub {set_diff($cur_diff+1);},
	    -background => 'lightgreen',
	    -width => $width
	)->grid(

	    $tk_action_frame->Button(
		-text => "Previous",
		-command => sub {set_diff($cur_diff-1);},
		-background => 'lightgreen',
		-width => $width
	    ),

	    $tk_action_frame->Button(
		-text => "First",
		-command => sub {set_diff(0);},
		-background => 'lightgreen',
		-width => $width
	    ),

	    $tk_action_frame->Button(
		-text => "Last",
		-command => sub {set_diff($#diff_info);},
		-background => 'lightgreen',
		-width => $width
	    ),
	    -sticky => "nsew"
	);

    if (defined($merge)) {
        $tk_action_frame->Button(
                -text => "Left",
                -command => \&cb_use_left,
                -background => 'cyan',
                -width => $width
            )->grid(
                $tk_action_frame->Button(
                    -text => "Right",
                    -command => \&cb_use_right,
                    -background => 'cyan',
                    -width => $width
                ),

                $tk_action_frame->Button(
                    -text => "Left then Right",
                    -command => \&cb_use_left_right,
                    -background => 'cyan',
                    -width => $width
                ),

                $tk_action_frame->Button(
                    -text => "Right then Left",
                    -command => \&cb_use_right_left,
                    -background => 'cyan',
                    -width => $width
                ),
                -sticky => "nsew"
        );
    }
    # Loop over each widget adding help
    foreach my $cur_widget ($tk_action_frame->gridSlaves())
    {
	# The label of the button
	my $label = $cur_widget->cget("-text");

	# The help message
	my $msg = $help{$label};

	if (not defined($msg))
	{
	    die("Unable to find help for $label");
	}
	$tk_mw->Balloon(
		-background => 'yellow'
	    )->attach(
		$cur_widget, 
		-msg => $msg
	    );
    }
}

################################################################
# Create_text_window($frame, $name, $scroll, $adjustor)
#
# Create a text frame, add adjustor if needed
################################################################
sub create_text_window($$$$)
{
    my $tk_text_frame = shift;  # Our parent
    my $name = shift;           # The name of this window
    my $scroll = shift;         # Where to put the scroll bars
    my $adjuster = shift;       # Add adjuster?

    my $tk_widget = $tk_text_frame->Scrolled(
		    'ROText',
		    -scrollbars => 'osoe',
		    -width => 30,
		    -background => 'white',
		    -wrap => 'none'
		)->pack(
		    -side => 'left',
		    -fill => 'both',
		    -expand => 1);

    if ($adjuster) {
        $tk_text_frame->Adjuster(
                )->packAfter(
                    $tk_widget, -side => 'left'
                );
    }
    return ($tk_widget);
}
################################################################
# build_gui -- Create the gui
################################################################
sub build_gui()
{
    if (defined($merge)) {
        # The main window
        $tk_mw = MainWindow->new( -title => "Merge");
    } else {
        # The main window
        $tk_mw = MainWindow->new( -title => "Diff");
    }

    # The frame to hold the action buttons at the top
    my $tk_button_frame = $tk_mw->Frame(
	    )->pack(
		-side => "top",
		-expand => 0,
		-fill => "x");

    # The exit button
    my $tk_exit = $tk_button_frame->Button(
	    -text => "Exit",
	    -command => \&cb_exit,
	    -background => 'white'
	)->pack(
	    -side => 'left'
	);

    $tk_mw->Balloon(
	    -background => 'yellow'
	)->attach(
	    $tk_exit, 
	    -msg => "Write merged file and exit"
	);

    # The frame to use to select text and such
    my $tk_select_frame = $tk_button_frame->Frame(
	    )->pack(
		-side => 'left'
	    );
    create_select_buttons($tk_select_frame);

    # The button to abort the code
    my $tk_abort = $tk_button_frame->Button(
	    -text => "Abort",
	    -command => \&cb_abort,
	    -background => 'pink'
	)->pack(
	    -side => 'left'
	);

    $tk_mw->Balloon(
	    -background => 'yellow'
	)->attach(
	    $tk_abort, 
	    -msg => "Exit without writing the output file"
	);
    # The text frame to hold the mess at the bottom
    my $tk_text_frame = $tk_mw->Frame(
	    )->pack(
		-side => 'top',
		-fill => 'both',
		-expand => 1
	    );

    # Text box for the original text
    my $tk_original;

    if (not defined ($opt_n))
    {
        $tk_original = create_text_window($tk_text_frame, "Ancestor: $opt_n", "os", 1);
    }
    my $tk_file1 = create_text_window($tk_text_frame, "File1: $file1", "os", 1);
    my $tk_file2 = create_text_window($tk_text_frame, "File2: $file2", 
        defined($merge) ? "osow" : "os",
        defined($merge) ? 1 : 0);


    $tk_scroll = $tk_file2->Subwidget('yscrollbar');

    %tk_text = (
            file1 => $tk_file1->Subwidget('rotext'),
            file2 => $tk_file2->Subwidget('rotext'),
        );

    if (defined($merge)) {
        # The window containing the merged text
        my $tk_merge = create_text_window($tk_text_frame, "Ancestor: $opt_n", "osow", 0);

        $tk_scroll = $tk_merge->Subwidget('yscrollbar');
        $tk_text{merge} = $tk_merge->Subwidget('rotext');
    }

    if (defined($tk_original))
    {
	$tk_text{original} = $tk_original->Subwidget('rotext');
    }

    $tk_scroll->configure(
	    -command => sub 
		{
		    foreach my $cur_text (values %tk_text)
		    {
		        $cur_text->yview(@_);
		    }
		}
	    );
    foreach my $cur_text (values %tk_text)
    {
	$cur_text->configure(
#		-yscrollcommand => sub {scroll_all($tk_scroll, $cur_text, @_); }
		-yscrollcommand => [\&scroll_all, $tk_scroll, $cur_text, @_]
	    );

    }
    # 
    # We define the tag
    #	diff -- The difference for the current text window
    #
    foreach my $cur_text (keys %tk_text)
    {
	$tk_text{$cur_text}->tagConfigure('diff',
		-background => $colors{$cur_text});
	$tk_text{$cur_text}->tagConfigure('cur_diff',
		-bgstipple => "gray50");
	$tk_text{$cur_text}->tagBind('diff', "<Button-1>", \&cb_click);
    }

    if (defined($merge)) {
        #
        # For merge we define the tags
        #	diff_<type> -- Difference type for each possible source
        #
        foreach my $cur_text (keys %tk_text)
        {
            $tk_text{merge}->tagConfigure("diff_$cur_text",
                    -background => $colors{$cur_text});
            $tk_text{merge}->tagBind("diff_$cur_text", "<Button-1>", \&cb_click);
        }
    }
    ##TODO: Horizontal scrolling (with and without locking)
}

################################################################
# usage -- Tell people how to use us
################################################################
sub usage()
{
    print STDERR "Usage is $0 [-n] -a <ancestor> -o <output> <file1> <file2>\n";
    exit (8);
}
################################################################
# find_diff3 -- Locate the ccm_dff3 command
################################################################
sub find_diff3()
{
    if (-f "C:\\")
    {
	my @path_words = split /;/, $ENV{PATH};
	foreach (@path_words)
	{
	    $diff3_cmd = "$_\\util\\ccm_dff3";
	    if (-f $diff3_cmd)
	    {
		return;
	    }
	}
    }
    else
    {
	my @path_words = split /:/, $ENV{PATH};
	foreach (@path_words)
	{
	    $diff3_cmd = "$_/util/ccm_dff3";
	    if (-f $diff3_cmd)
	    {
		return;
	    }
	}
    }
    die ("Could not find the ccm_dff3 command");
}
find_diff3();
#
# Figure out which files we have
#	-a ancestor
#	-o out_file
#	-n no ancestor pane
#       -D debug
my $status = getopts('a:o:nd');
if ($status == 0)
{
    usage();
}
if (not defined($opt_a))
{
    usage();
}
if ($#ARGV != 1)
{
    usage();
}
$original = $opt_a;
$file1 = $ARGV[0];
$file2 = $ARGV[1];
$merge = $opt_o;

compute_diff_info();	# Compute the difference information

# Read files and store the lines in the files
my @original_lines = read_file($original);
my @file1_lines = read_file($file1);
my @file2_lines = read_file($file2);
my @merge_lines = @file1_lines;	

foreach my $cur_diff (@diff_info)
{
    $cur_diff->{show}{merge} = { %{$cur_diff->{show}{file1}} };
    $cur_diff->{result} = "left";
}

%lines = (
    original => \@original_lines,
    file1 => \@file1_lines,
    file2 => \@file2_lines,
    merge => \@merge_lines
);

#TODO: Make part of debugging
#display_text_diff();
build_gui();
if (defined ($tk_text{original}))
{
    fill_box("original");
}
fill_box("file1");
fill_box("file2");
if (defined($merge)) {
    fill_box("merge");
}
set_diff(0);

MainLoop();

